// Copyright (c) 2017 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SPIRV_TOOLS_COMP_MARKV_MODEL_FACTORY_H_
#define SPIRV_TOOLS_COMP_MARKV_MODEL_FACTORY_H_

#include <memory>

#include "source/comp/markv_model.h"

namespace spvtools {

enum MarkvModelType {
  kMarkvModelUnknown = 0,
  kMarkvModelShaderLite,
  kMarkvModelShaderMid,
  kMarkvModelShaderMax,
};

std::unique_ptr<MarkvModel> CreateMarkvModel(MarkvModelType type);

}  // namespace spvtools

#endif  // SPIRV_TOOLS_COMP_MARKV_MODEL_FACTORY_H_
